from typing import List, Literal

from airflow_pydantic import BaseModel, BashCommands
from pydantic import Field

__all__ = (
    "clone_repo",
    "GitRepo",
    "PipLibrary",
    "LibraryList",
    "Library",
    "Tool",
)


Tool = Literal["pip", "uv"]


def clone_repo(
    name: str,
    repo: str,
    branch: str = "main",
    *,
    clean: bool = False,
    install: bool = True,
    install_deps: bool = False,
    tool: Tool = "pip",
    dir: str = "",
):
    cmds = [
        f"[[ -d {name} ]] || git clone {repo}",
        f"pushd {name}",
        "git stash",
    ]
    if clean:
        cmds.append("git clean -fdx")
    cmds.extend(
        [
            "git fetch --all --force",
            f"git checkout {branch}",
            f"git reset origin/{branch} --hard",
        ]
    )
    if dir:
        cmds.insert(0, f"cd {dir}")
        cmds.insert(0, f"mkdir -p {dir}")
    if install:
        tool = "uv pip" if tool == "uv" else "pip"
        install_deps_flag = "" if install_deps else "--no-deps "
        cmd = f"{tool} install {install_deps_flag}-e ."
        cmds.append(f"{cmd}")
    return BashCommands(commands=cmds)._serialize()


class GitRepo(BaseModel):
    name: str
    repo: str
    branch: str = "main"

    clean: bool = False
    install: bool = True
    install_deps: bool = False
    tool: Tool = "pip"
    dir: str = ""

    def clone(self):
        return clone_repo(
            name=self.name,
            repo=self.repo,
            branch=self.branch,
            clean=self.clean,
            install=self.install,
            install_deps=self.install_deps,
            tool=self.tool,
            dir=self.dir,
        )


class PipLibrary(BaseModel):
    name: str

    version_constraint: str = ""
    install_deps: bool = False
    tool: Tool = "pip"
    dir: str = ""

    def install(self):
        tool = "uv pip" if self.tool == "uv" else "pip"
        install_deps_flag = "" if self.install_deps else "--no-deps "
        install_dir_flag = "" if not self.dir else f"--target {self.dir} "
        return BashCommands(commands=[f'{tool} install {install_deps_flag}{install_dir_flag}"{self.name}{self.version_constraint}"'])._serialize()


class LibraryList(BaseModel):
    pip: List[PipLibrary] = Field(default_factory=list)
    git: List[GitRepo] = Field(default_factory=list)


# Alias
Library = LibraryList



from typing import List, Optional, Type

from airflow_pydantic import BashTask, BashTaskArgs, ImportPath, SSHTask, SSHTaskArgs
from airflow_pydantic.airflow import PythonOperator
from pydantic import Field, field_validator

from .model import GitRepo, PipLibrary

__all__ = (
    "LibraryOperatorBase",
    "InstallLibraryOperator",
    "InstallLibrarySSHOperator",
    "LibraryListTaskArgs",
    "LibraryListSSHTaskArgs",
    "LibraryListTask",
    "LibraryListSSHTask",
)


class LibraryOperatorBase(PythonOperator):
    _first_command: "PythonOperator"
    _last_command: "PythonOperator"

    def __init__(
        self,
        task_id: str,
        pip: List[PipLibrary],
        git: List[GitRepo],
        operator,
        operator_command_arg,
        parallel: Optional[bool] = False,
        command_prefix: Optional[str] = "",
        **kwargs,
    ):
        self.pip = pip
        self.git = git
        self.operator = operator
        self.parallel = parallel
        if command_prefix:
            self.command_prefix = f"{command_prefix}\n"
        else:
            self.command_prefix = ""

        # Print info, run each git library, run each pip library
        info = []
        for lib in self.git:
            info.append(f"Git library: {lib}")
        for lib in self.pip:
            info.append(f"Pip library: {lib}")

        # Initialize self
        super().__init__(
            task_id=f"{task_id}-info",
            python_callable=lambda: print("\n".join(info)),
            dag=kwargs.get("dag", None),
        )

        # Setup deps
        last_command = self
        for git_repo in self.git:
            git_repo_clone_op = self.operator(
                task_id=f"{task_id}-{git_repo.name}-clone",
                **{operator_command_arg: f"{self.command_prefix}{git_repo.clone()}"},
                **kwargs,
            )
            # NOTE: overloading shift so use parent's
            PythonOperator.set_downstream(last_command, git_repo_clone_op)
            last_command = self if self.parallel else git_repo_clone_op

        for pip_lib in self.pip:
            pip_lib_install_op = self.operator(
                task_id=f"{task_id}-{pip_lib.name}-install",
                **{operator_command_arg: f"{self.command_prefix}{pip_lib.install()}"},
                **kwargs,
            )
            # NOTE: overloading shift so use parent's
            PythonOperator.set_downstream(last_command, pip_lib_install_op)
            last_command = self if self.parallel else pip_lib_install_op

        # Overload sequence
        self._first_command = self
        self._last_command = last_command

    def __lshift__(self, other):
        """<<"""
        self._first_command << other
        return self

    def __rshift__(self, other):
        """>>"""
        self._last_command >> other
        return self

    def set_upstream(self, other):
        self._first_command.set_upstream(other)
        return self

    def set_downstream(self, other):
        self._last_command.set_downstream(other)


class InstallLibraryOperator(LibraryOperatorBase):
    def __init__(
        self, task_id: str, pip: List[PipLibrary], git: List[GitRepo], parallel: Optional[bool] = False, command_prefix: Optional[str] = "", **kwargs
    ):
        from airflow.providers.standard.operators.bash import BashOperator

        obj = LibraryListTaskArgs.model_validate({"pip": pip, "git": git, "parallel": parallel, "command_prefix": command_prefix})

        super().__init__(
            task_id=task_id,
            pip=obj.pip,
            git=obj.git,
            parallel=obj.parallel,
            command_prefix=obj.command_prefix,
            operator=BashOperator,
            operator_command_arg="bash_command",
            **kwargs,
        )


class InstallLibrarySSHOperator(LibraryOperatorBase):
    def __init__(
        self, task_id: str, pip: List[PipLibrary], git: List[GitRepo], parallel: Optional[bool] = False, command_prefix: Optional[str] = "", **kwargs
    ):
        from airflow.providers.ssh.operators.ssh import SSHOperator

        obj = LibraryListSSHTaskArgs.model_validate({"pip": pip, "git": git, "parallel": parallel, "command_prefix": command_prefix})

        super().__init__(
            task_id=task_id,
            pip=obj.pip,
            git=obj.git,
            parallel=obj.parallel,
            command_prefix=obj.command_prefix,
            operator=SSHOperator,
            operator_command_arg="command",
            **kwargs,
        )


class _LibraryCommonArgs:
    pip: List[PipLibrary] = Field(default_factory=list)
    git: List[GitRepo] = Field(default_factory=list)

    parallel: Optional[bool] = Field(default=False)
    command_prefix: Optional[str] = Field(default="")


class LibraryListTaskArgs(BashTaskArgs, _LibraryCommonArgs): ...


class LibraryListSSHTaskArgs(SSHTaskArgs, _LibraryCommonArgs): ...


class LibraryListTask(BashTask, LibraryListTaskArgs):
    operator: ImportPath = Field(default="airflow_common_operators.InstallLibraryOperator", validate_default=True)

    @field_validator("operator")
    @classmethod
    def validate_operator(cls, v: Type) -> Type:
        if not isinstance(v, Type) and issubclass(v, InstallLibraryOperator):
            raise ValueError(f"operator must be 'airflow_common_operators.InstallLibraryOperator', got: {v}")
        return v


class LibraryListSSHTask(SSHTask, LibraryListTaskArgs):
    operator: ImportPath = Field(default="airflow_common_operators.InstallLibrarySSHOperator", validate_default=True)

    @field_validator("operator")
    @classmethod
    def validate_operator(cls, v: Type) -> Type:
        if not isinstance(v, Type) and issubclass(v, InstallLibrarySSHOperator):
            raise ValueError(f"operator must be 'airflow_common_operators.InstallLibrarySSHOperator', got: {v}")
        return v



from typing import TYPE_CHECKING, List, Optional

from airflow_pydantic import fail, pass_, ping
from airflow_pydantic.airflow import PythonOperator

if TYPE_CHECKING:
    from airflow.models.baseoperator import BaseOperator
    from airflow.models.dag import DAG

__all__ = ("all_success_any_failure", "if_booted_do")


def all_success_any_failure(
    *,
    task_id: str,
    tasks: List["BaseOperator"],
    dag: "DAG",
    queue: Optional[str] = None,
):
    any_ssh_failure = PythonOperator(
        task_id=f"{task_id}-any-failure",
        python_callable=fail,
        queue=queue,
        trigger_rule="one_failed",
        dag=dag,
    )
    all_ssh_success = PythonOperator(
        task_id=f"{task_id}-all-success",
        python_callable=pass_,
        queue=queue,
        trigger_rule="none_failed",
        dag=dag,
    )

    for task in tasks:
        task >> any_ssh_failure
        task >> all_ssh_success

    return any_ssh_failure, all_ssh_success


def if_booted_do(task_id: str, host: str, task: "BaseOperator", **check_operators_kwargs) -> "BaseOperator":
    check_if_booted = PythonOperator(
        task_id=f"{task_id}-check-if-booted-{host}",
        python_callable=ping(host),
        **check_operators_kwargs,
    )

    if isinstance(task, list):
        for task_instance in task:
            check_if_booted >> task_instance
    else:
        check_if_booted >> task
    return task



from airflow_common_operators import GitRepo, PipLibrary, clone_repo


class TestLibrary:
    def test_pip(self):
        p = PipLibrary(name="tmp", version_constraint="<5", install_deps=False, tool="uv")
        assert p.install() == "bash -lc 'set -ex\nuv pip install --no-deps \"tmp<5\"'"
        p = PipLibrary(name="tmp", version_constraint="", install_deps=True, tool="pip", dir="/tmp")
        assert p.install() == "bash -lc 'set -ex\npip install --target /tmp \"tmp\"'"
        p = PipLibrary(name="tmp", version_constraint=">=3.2", install_deps=True, tool="pip", dir="/tmp")
        assert p.install() == "bash -lc 'set -ex\npip install --target /tmp \"tmp>=3.2\"'"

    def test_git(self):
        g = GitRepo(name="tmp", repo="tmp", branch="main", clean=True, install=True, install_deps=False, tool="uv")
        assert (
            g.clone()
            == "bash -lc 'set -ex\n[[ -d tmp ]] || git clone tmp\npushd tmp\ngit stash\ngit clean -fdx\ngit fetch --all --force\ngit checkout main\ngit reset origin/main --hard\nuv pip install --no-deps -e .'"
        )
        g = GitRepo(name="tmp", repo="tmp", branch="main", install=True, install_deps=True, tool="pip", dir="/tmp")
        assert (
            g.clone()
            == "bash -lc 'set -ex\nmkdir -p /tmp\ncd /tmp\n[[ -d tmp ]] || git clone tmp\npushd tmp\ngit stash\ngit fetch --all --force\ngit checkout main\ngit reset origin/main --hard\npip install -e .'"
        )
        assert (
            clone_repo(name="tmp", repo="tmp", branch="main", install=False, install_deps=True, tool="uv")
            == "bash -lc 'set -ex\n[[ -d tmp ]] || git clone tmp\npushd tmp\ngit stash\ngit fetch --all --force\ngit checkout main\ngit reset origin/main --hard'"
        )



command = "sudo journalctl --vacuum-time=2d"
"""
echo 'sudo systemctl stop airflow-celery-worker' | at now + 1 minute
echo 'sudo systemctl stop log2ram' | at now + 1 minute
echo 'sudo reboot now' | at now + 2 minutes
"""

"""
echo 'sudo systemctl stop airflow-celery-worker' | at now + 1 minute
echo 'sudo reboot now' | at now + 2 minutes
"""

"""
echo 'sudo lunchy stop timkpaine.airflow-celery-worker' | at now + 1 minute
echo 'sudo reboot now' | at now + 2 minutes
"""

"""
echo 'sudo systemctl stop airflow-webserver' | at now + 5 minute
echo 'sudo systemctl stop airflow-scheduler' | at now + 5 minute
echo 'sudo systemctl stop airflow-triggerer' | at now + 5 minute
echo 'sudo systemctl stop airflow-celery-flower' | at now + 5 minute
echo 'sudo systemctl stop airflow-celery-worker' | at now + 5 minute
echo 'sudo systemctl stop airflow-posgresql' | at now + 5 minute
echo 'sudo systemctl stop redis' | at now + 5 minute
echo 'sudo reboot now' | at now + 6 minutes
"""



from airflow_config import load_config

RENDERED_DAG = """# Generated by airflow-config
from datetime import datetime

from airflow.models import DAG

from airflow_common_operators.infra.task import JournalctlClean

with DAG(
    description="this is an example dag",
    schedule="0 3 * * *",
    start_date=datetime.fromisoformat("2023-01-01T00:00:00"),
    dag_id="clean-journalctl",
    default_args={},
) as dag:
    clean = JournalctlClean(sudo=True, days=2, task_id="clean", dag=dag)
"""


class TestConfig:
    def test_load_config_hydra(self):
        config = load_config(config_name="config")
        assert config
        assert "libraries" in config.extensions
        assert len(config.extensions["libraries"].pip) == 3
        assert len(config.extensions["libraries"].git) == 3

    def test_render_and_instantiate(self):
        config = load_config(config_name="config")
        assert config.dags["clean-journalctl"].render() == RENDERED_DAG
        exec(config.dags["clean-journalctl"].render())



from airflow_pydantic import fail, pass_, skip
from airflow_pydantic.airflow import PythonOperator

__all__ = (
    "Skip",
    "Fail",
    "Pass",
)


def Skip(task_id: str, **kwargs) -> PythonOperator:
    return PythonOperator(task_id=task_id, python_callable=skip, **kwargs)


def Fail(task_id: str, **kwargs) -> PythonOperator:
    return PythonOperator(task_id=task_id, python_callable=fail, **kwargs)


def Pass(task_id: str, **kwargs) -> PythonOperator:
    return PythonOperator(task_id=task_id, python_callable=pass_, **kwargs)



from typing import Optional, Type

from airflow_pydantic import BashTask, BashTaskArgs, CallablePath, SSHTask, SSHTaskArgs
from airflow_pydantic.airflow import BashOperator, SSHOperator
from pydantic import Field, field_validator

__all__ = (
    "JournalctlClean",
    "JournalctlCleanSSH",
    "JournalctlCleanOperator",
    "JournalctlCleanOperatorArgs",
    "JournalctlCleanSSHOperatorArgs",
    "JournalctlCleanSSHOperator",
    "JournalctlCleanTask",
    "JournalctlCleanTaskArgs",
    "JournalctlCleanSSHTask",
    "JournalctlCleanSSHTaskArgs",
)


def clean_journalctl(sudo: Optional[bool] = True, days: Optional[int] = 2):
    days = days or 2
    cmd = f"sudo journalctl --vacuum-time={days}d" if sudo else f"journalctl --vacuum-time={days}d"
    return cmd


class JournalctlClean(BashOperator):
    def __init__(self, sudo: Optional[bool] = True, days: Optional[int] = 2, **kwargs):
        if "bash_command" in kwargs:
            raise ValueError("JournalctlClean does not accept 'bash_command' as an argument.")
        super().__init__(bash_command=clean_journalctl(sudo=sudo, days=days), **kwargs)


class JournalctlCleanSSH(SSHOperator):
    def __init__(self, sudo: Optional[bool] = True, days: Optional[int] = 2, **kwargs):
        if "command" in kwargs:
            raise ValueError("JournalctlCleanSSH does not accept 'command' as an argument.")
        super().__init__(command=clean_journalctl(sudo=sudo, days=days), **kwargs)


class JournalctlCleanTaskArgs(BashTaskArgs):
    sudo: Optional[bool] = Field(default=True)
    days: Optional[int] = Field(default=2)


class JournalctlCleanSSHTaskArgs(SSHTaskArgs):
    sudo: Optional[bool] = Field(default=True)
    days: Optional[int] = Field(default=2)


# Alias
JournalctlCleanOperatorArgs = JournalctlCleanTaskArgs
JournalctlCleanSSHOperatorArgs = JournalctlCleanSSHTaskArgs


class JournalctlCleanTask(BashTask, JournalctlCleanTaskArgs):
    operator: CallablePath = Field(default="airflow_common_operators.JournalctlClean", validate_default=True)

    @field_validator("operator")
    @classmethod
    def validate_operator(cls, v: Type) -> Type:
        if v is not JournalctlClean:
            raise ValueError(f"operator must be 'airflow_common_operators.JournalctlClean', got: {v}")
        return v


class JournalctlCleanSSHTask(SSHTask, JournalctlCleanSSHTaskArgs):
    operator: CallablePath = Field(default="airflow_common_operators.JournalctlCleanSSH", validate_default=True)

    @field_validator("operator")
    @classmethod
    def validate_operator(cls, v: Type) -> Type:
        if v is not JournalctlCleanSSH:
            raise ValueError(f"operator must be 'airflow_common_operators.JournalctlCleanSSH', got: {v}")
        return v


# Alias
JournalctlCleanOperator = JournalctlCleanTask
JournalctlCleanSSHOperator = JournalctlCleanSSHTask



from airflow_config import load_config

RENDERED_DAG = """# Generated by airflow-config
from datetime import datetime

from airflow.models import DAG
from airflow.providers.standard.operators.bash import BashOperator

from airflow_common_operators.library.task import InstallLibrarySSHOperator

with DAG(
    description="this is an example dag",
    schedule="0 3 * * *",
    start_date=datetime.fromisoformat("2023-01-01T00:00:00"),
    dag_id="setup-env",
    default_args={},
) as dag:
    first = BashOperator(bash_command="echo '1'", task_id="first", dag=dag)
    run = InstallLibrarySSHOperator(
        pip=[
            {"name": "lib1", "version_constraint": ">=1.0"},
            {"name": "lib2", "version_constraint": "<=1.0"},
            {"name": "lib3", "version_constraint": "!=1.0"},
        ],
        git=[{"name": "git1", "repo": "test"}, {"name": "git2", "repo": "test", "branch": "develop"}, {"name": "git3", "repo": "test"}],
        command_prefix='echo "test"',
        task_id="run",
        dag=dag,
    )
    last = BashOperator(bash_command="echo '2'", task_id="last", dag=dag)
    first >> run
    run >> last
"""


class TestConfig:
    def test_load_config_hydra(self):
        config = load_config(config_name="config")
        assert config
        assert "libraries" in config.extensions
        assert len(config.extensions["libraries"].pip) == 3
        assert len(config.extensions["libraries"].git) == 3

    def test_render_and_instantiate(self):
        config = load_config(config_name="config")
        assert config.dags["setup-env"].render() == RENDERED_DAG
        exec(config.dags["setup-env"].render())



from unittest.mock import MagicMock, call, patch

import pytest
from airflow_pydantic import fail, pass_

from airflow_common_operators import all_success_any_failure, if_booted_do


class TestHelpers:
    @pytest.mark.parametrize("make_list", [True, False])
    def test_if_booted_do(self, make_list):
        with (
            patch("airflow_common_operators.airflow.topology.PythonOperator") as mock_python_operator,
            patch("airflow_common_operators.airflow.topology.ping") as mock_ping,
        ):
            if make_list:
                task = if_booted_do(task_id="task", host="host", task=[MagicMock()])
            else:
                task = if_booted_do(task_id="task", host="host", task=MagicMock())
            mock_ping.assert_called_once_with("host")
            mock_python_operator.assert_called_once_with(
                task_id="task-check-if-booted-host",
                python_callable=mock_ping.return_value,
            )
            if make_list:
                mock_python_operator.return_value.__rshift__.assert_called_once_with(task[0])
            else:
                mock_python_operator.return_value.__rshift__.assert_called_once_with(task)

    def test_all_success_any_failure(self):
        with patch("airflow_common_operators.airflow.topology.PythonOperator") as mock_python_operator:
            task = MagicMock()
            dag = MagicMock()

            any_ssh_failure, all_ssh_success = all_success_any_failure(
                task_id="task",
                tasks=[task],
                dag=dag,
                queue="blerg",
            )

            mock_python_operator.assert_has_calls(
                [
                    call(task_id="task-any-failure", python_callable=fail, queue="blerg", trigger_rule="one_failed", dag=dag),
                    call(task_id="task-all-success", python_callable=pass_, queue="blerg", trigger_rule="none_failed", dag=dag),
                ]
            )
            task.__rshift__.assert_has_calls(
                [
                    call(any_ssh_failure),
                    call(all_ssh_success),
                ]
            )
